from pathlib import Path
from git import Repo, GitCommandError
import sys
import stat
import inspect
import os
from typing import List
import shutil

from .errors import PipelineExistsError, PipelineNotFoundError

class Pipeline:
    def __init__(self, path:Path) -> None:
        self.path = path
        self.repo = Repo(path)
        self.name = self.path.name
    
    @property
    def version(self):
        try:
            version = self.repo.git.describe(['--tags','--exact-match'])
        except GitCommandError:
            version = None
        return version

    @property
    def executable(self):
        pipeline_bin_dir = self.path / 'bin'
        name = self.name
        if sys.platform.startswith('win'):
            name += '.exe'
        return pipeline_bin_dir / name
        


class Nest:
    """Class for managing Snk pipelines"""

    def __init__(self, snk_home: Path = None, bin_dir: Path = None) -> None:
        self.python_interpreter_path = Path(sys.executable) # needs to be the same python that has snk
        
        if not snk_home:
            snk_home = self.python_interpreter_path.parent.parent / 'snk'
            if not os.access(snk_home, os.W_OK):
                user_home_path = Path('~').expanduser()
                snk_home = user_home_path / ".local" / 'snk'
        if not bin_dir:
            bin_dir = self.python_interpreter_path.parent
            if not os.access(bin_dir, os.W_OK):
                user_home_path = Path('~').expanduser()
                bin_dir = user_home_path / ".local" / 'bin'

        self.snk_home = Path(snk_home).absolute()
        self.pipelines_dir = self.snk_home / "pipelines"
        self.bin_dir = Path(bin_dir).absolute()
        
        # Create dirs
        self.snk_home.mkdir(parents=True, exist_ok=True)
        self.pipelines_dir.mkdir(parents=True, exist_ok=True)
        self.bin_dir.mkdir(parents=True, exist_ok=True)

    def _check_repo_url_format(self, repo):
        if not repo.endswith('.git'):
            raise ValueError('Repo url must end in .git')

    def install(self, repo_url: str, name=None, version=None, config=None) -> Pipeline:
        """
        Install a Snakemake pipeline as a CLI. 
        They must be standards compliant, public, Snakemake workflows.
        https://snakemake.github.io/snakemake-workflow-catalog/
        """
        self._check_repo_url_format(repo_url)
        if not name:
            name = self._get_name_from_git_url(repo_url)
        if name in os.listdir(self.pipelines_dir):
            raise PipelineExistsError(f"Pipeline '{name}' already exists in {self.pipelines_dir}")
        if name in os.listdir(self.bin_dir):
            raise PipelineExistsError(f"Pipeline '{name}' already exists in {self.bin_dir}")
        repo = self.download(repo_url, name, tag_name=version)
        try:
            pipeline = Pipeline(path=Path(repo.git_dir).parent)
            pipeline_executable = self.create_package(pipeline.path)
            self.link_pipeline_executable_to_bin(pipeline_executable)
            if config:
                self.copy_nonstandard_config(pipeline.path, config)
            self._confirm_installation(name)
        except Exception as e:
            # remove any half completed steps 
            to_remove = self.get_paths_to_delete(name)
            self.delete_paths(to_remove)
            raise e
        return pipeline

    def copy_nonstandard_config(self, pipeline_dir: Path, config_path: Path):
        config_dir = pipeline_dir / 'config'
        config_dir.mkdir()
        shutil.copyfile(pipeline_dir / config_path, config_dir / 'config.yaml')

    def get_paths_to_delete(self, pipeline_name: str) -> List[Path]:
        to_delete = []
        
        # remove repo 
        pipeline_dir = self.pipelines_dir / pipeline_name
        if pipeline_dir.exists() and pipeline_dir.is_dir():
            to_delete.append(pipeline_dir)
        else:
            raise PipelineNotFoundError(f'Could not find pipeline: {pipeline_name}')

        # remove link
        pipeline_bin_executable = self.bin_dir / pipeline_name
        if pipeline_bin_executable.exists() and pipeline_bin_executable.is_symlink():
            if str(pipeline_dir) in str(pipeline_bin_executable.readlink()):
                to_delete.append(pipeline_bin_executable)
        
        return to_delete

    def delete_paths(self, files: List[Path]):
        # check that the files are in self.pipelines_dir
        # i.e. if it is a symlink read the link and check
        for path in files:
            if path.is_dir():
                assert str(self.snk_home) in str(path), "Cannot delete folders outside of SNK_HOME"
                shutil.rmtree(path)
            elif path.is_symlink():
                assert str(self.snk_home) in str(path.readlink()), "Cannot delete files outside of SNK_HOME"
                path.unlink()
            else:
                raise TypeError("Invalid file type")

    def uninstall(self, name: str, force: bool = False) -> bool:
        to_remove = self.get_paths_to_delete(name)
        if force:
            proceed = True
        else:
            print(f"Uninstalling {name}")
            print("  Would remove:")
            for p in to_remove:
                print(f"    {p}{'/*' if p.is_dir() else ''}")
            ans = input("Proceed (Y/n)? ")
            proceed = ans.lower() in ['y', 'yes']
        if not proceed:
            return False
        self.delete_paths(to_remove)
        return True 

    def _confirm_installation(self, name: str):
        pipeline_dir = self.pipelines_dir / name
        assert pipeline_dir.exists()
        pipelines = [p.name.split('.')[0] for p in self.bin_dir.glob('*')]
        assert name in pipelines

    def _get_name_from_git_url(self, git_url: str):
        return git_url.split('/')[-1].split('.')[0]

    @property
    def pipelines(self):
        return [Pipeline(pipeline_dir.resolve()) for pipeline_dir in self.pipelines_dir.glob('*') if pipeline_dir.is_dir()]

    def download(self, repo_url: str, name: str, tag_name: str = None) -> Repo:
        """Pull the repo"""
        location  = self.pipelines_dir / name
        options = ['--depth 1', '--single-branch']
        if tag_name:
            options.append(f'--branch {tag_name}')
        try:
            repo = Repo.clone_from(repo_url, location, multi_options=options)
            repo.git.checkout(tag_name)
        except GitCommandError as e:
            if "destination path" in e.stderr:
                raise PipelineExistsError(f"Pipeline '{name}' already exists in {self.pipelines_dir}")
            elif f"Remote branch {tag_name}" in e.stderr:
                raise PipelineNotFoundError(f"Pipeline tag '{tag_name}' not found")
            elif "not found" in e.stderr:
                raise PipelineNotFoundError(f"Pipeline repository '{repo_url}' not found")
            raise e
        return repo

    def create_package(self, pipeline_dir: Path) -> Path:
        """Convert a SnakeMake pipeline repo into a snk CLI"""
        self.validate_SnakeMake_repo(pipeline_dir)
        
        template = inspect.cleandoc(f"""
            #!/bin/sh
            '''exec' "{self.python_interpreter_path}" "$0" "$@"
            ' '''
            # -*- coding: utf-8 -*-
            import re
            import sys
            from snk.pipeline import cli
            if __name__ == "__main__":
                sys.argv[0] = re.sub(r'(-script\.pyw|\.exe)?$', '', sys.argv[0])
                sys.exit(cli("{pipeline_dir}"))
                
        """)

        pipeline_bin_dir = pipeline_dir / 'bin'
        pipeline_bin_dir.mkdir(exist_ok=True)

        name = pipeline_dir.name

        if sys.platform.startswith('win'):
            name += '.exe'
        
        pipeline_executable = pipeline_bin_dir / name

        with open(pipeline_executable, 'w') as f:
            f.write(template)

        pipeline_executable.chmod(pipeline_executable.stat().st_mode | stat.S_IEXEC)

        return pipeline_executable

    def link_pipeline_executable_to_bin(self, pipeline_executable_path: Path):
        name = pipeline_executable_path.name
        os.symlink(pipeline_executable_path.absolute(), self.bin_dir / name)
        return self.bin_dir / name

    def validate_SnakeMake_repo(self, repo: Repo):
        print("Skipping validation!")
        pass


import typer
from pathlib import Path
from typing import Optional, List
import sys
import collections
if sys.version_info.major == 3 and sys.version_info.minor >= 10:
    from collections.abc import MutableMapping
else:
    from collections import MutableMapping

import snakemake
from rich.console import Console
from rich.syntax import Syntax
from art import text2art
import subprocess

from .utils import add_dynamic_options
from snk.nest import Pipeline

def _print_snakemake_help(value: bool):
    if value:
        snakemake.main("-h")

def create_logo(name):
    logo = text2art(name, font="small")        
    doc  = f"""\b{logo}\bA Snakemake pipeline CLI generated with snk"""
    return doc

def convert_key_to_samkemake_format(key, value):
    resultDict = dict()
    parts = key.split(":")
    d = resultDict
    for part in parts[:-1]:
        if part not in d:
            d[part] = dict()
        d = d[part]
    d[parts[-1]] = value
    return resultDict

def parse_config_args(args: List[str], options):
    names = [op['name'] for op in options]
    config = []
    parsed = []
    flag=None
    for arg in args:
        if flag:
            name = flag.lstrip('-')
            op = next(op for op in options if op['name'] == name)
            if ":" in op['original_key']:
                samkemake_format_config = convert_key_to_samkemake_format(op['original_key'], arg)
                name = list(samkemake_format_config.keys())[0]
                arg = samkemake_format_config[name]
            if arg != 'None':
                # skip null args
                config.append(f"{name}={arg}")
            flag=None
            continue
        if arg.startswith('-') and arg.lstrip('-') in names:
            flag = arg
            continue
        parsed.append(arg)
    parsed.extend(["--config", *config])
    return parsed

def get_config_from_pipeline_dir(pipeline_dir_path: Path):
    """"""
    config_path = pipeline_dir_path / 'config' / 'config.yaml'
    return config_path 

def flatten(d, parent_key='', sep=':'):
    items = []
    for k, v in d.items():
        new_key = parent_key + sep + k if parent_key else k
        if isinstance(v, MutableMapping):
            items.extend(flatten(v, new_key, sep=sep).items())
        else:
            items.append((new_key, v))
    return dict(items)

def load_options(pipeline_dir_path: Path):
    pipeline_config_path = get_config_from_pipeline_dir(pipeline_dir_path)
    if not pipeline_config_path.exists():
        return []
    config = snakemake.load_configfile(pipeline_config_path)
    flat_config = flatten(config)
    options = []
    catalog_config_path = pipeline_dir_path / '.snakemake-workflow-catalog.yml'
    snk_config_path = pipeline_dir_path / '.snk'
    if snk_config_path.exists():
        snk_config = snakemake.load_configfile(snk_config_path)
    elif catalog_config_path.exists():
        catalog_config = snakemake.load_configfile(catalog_config_path)
        snk_config = flatten(catalog_config.get('snk', {}))
    else:
        snk_config = {}
    snk_annotations = snk_config.get('annotations', {})
    for op in flat_config:
        name = op.replace(':', '_')
        help = snk_annotations.get(f"{op}:help")
        type = snk_annotations.get(f"{op}:type")
        options.append(
            {
                'name':name.replace('-', '_'),
                'original_key': op,
                'default': str(flat_config[op]),
                'help': help if help else "",
                'type': type if type else "str"
            }
        )
    return options

def pipeline_cli_factory(pipeline_dir_path: Path):
    app = typer.Typer()
    options = load_options(pipeline_dir_path)
    pipeline = Pipeline(path=pipeline_dir_path)
    def _print_pipline_version(value: bool):
        if value:
            typer.echo(pipeline.version)
            raise typer.Exit()

    def _print_pipline_path(value: bool):
        if value:
            typer.echo(pipeline.path)
            raise typer.Exit()

    @app.callback(invoke_without_command=True, context_settings={"help_option_names": ["-h", "--help"]})
    def callback(
            ctx: typer.Context, 
            version: Optional[bool] = typer.Option(None, '-v', '--version', help="Show the pipeline version.", is_eager=True, callback=_print_pipline_version, show_default=False),
            path: Optional[bool] = typer.Option(None, '-p', '--path', help="Show the pipeline path.", is_eager=True, callback=_print_pipline_path, show_default=False)
        ):
        if ctx.invoked_subcommand is None:
            typer.echo(f'{ctx.get_help()}')
    # dynamically create the logo
    callback.__doc__ = f"{create_logo(pipeline_dir_path.name)}"

    @app.command(help="Run the pipeline. All unrecognized arguments are parsed onto Snakemake to be used by the pipeline.", context_settings={"allow_extra_args": True, "ignore_unknown_options": True, "help_option_names": ["-h", "--help"]})
    @add_dynamic_options(options)
    def run(
            ctx: typer.Context,
            target: str = typer.Argument(None, help="File to generate. If None will run the pipeline 'all' rule."),
            configfile: Path = typer.Option(None, help="Path to config file. Overrides existing config and defaults.", exists=True, dir_okay=False),
            cores:  int = typer.Option(None, help="Set the number of cores to use. If None will use all cores."),
            verbose: Optional[bool] = typer.Option(False, "--verbose", "-v", help="Run pipeline in verbose mode.",),
            help_snakemake: Optional[bool] = typer.Option(
                False, "--help-snakemake", "-hs", help="Print the snakemake help and exit.", is_eager=True, callback=_print_snakemake_help, show_default=False
            ),
        ):
        args = []
        conda_prefix_dir = pipeline_dir_path / '.conda'
        if not cores:
            cores = 'all'
        args.extend([
            "--use-conda",
            f"--conda-prefix={conda_prefix_dir}",
            f"--cores={cores}",
        ])
        snakefile = pipeline_dir_path / 'workflow' / 'Snakefile'
        if not snakefile.exists():
            raise ValueError('Could not find Snakefile')
        else:
            args.append(f"--snakefile={snakefile}")
        
        if not configfile:
            configfile = get_config_from_pipeline_dir(pipeline_dir_path)
        args.append(f"--configfile={configfile}")

        
        # Set up conda frontend
        mamba_found = True
        try:
            subprocess.run(["mamba", "--version"], capture_output=True, check=True)
        except (subprocess.CalledProcessError, FileNotFoundError):
            mamba_found = False
        if not mamba_found:
            args.append("--conda-frontend=conda")
        
        typer.echo(create_logo(pipeline_dir_path.name))
        typer.echo()

        if verbose:
            args.insert(0, "--verbose")

        if target:
            args.append(target)
        args.extend(parse_config_args(ctx.args, options=options))

        if verbose:
            typer.secho(f"snakemake {' '.join(args)}\n", fg=typer.colors.MAGENTA)

        snakemake.main(args)
    
    @app.command(help="Access the pipeline configuration.")
    def config():
        config_path = pipeline_dir_path / 'config' / 'config.yaml'
        if not config_path.exists():
            typer.secho("Could not find config...", fg='red')
            raise typer.Exit(1)
        with open(config_path) as f:
            code = f.read()
            syntax = Syntax(code, 'yaml')
            console = Console()
            console.print(syntax)
    
    
    @app.command(help="Display information about current pipeline install.")
    def info():
        import json
        info_dict = {}
        info_dict['name'] = pipeline_dir_path.name
        info_dict['version'] = pipeline.version
        info_dict['pipeline_dir_path'] = str(pipeline_dir_path)
        typer.echo(json.dumps(info_dict, indent=2))
    
    @app.command(help="Access the pipeline conda environments.")
    def env(
        activate: str = typer.Option(None, help="Activate a given environment"),
    ):
        raise NotImplementedError
    
            
    return app
 




import pytest
import tempfile
from pathlib import Path
from snk import Nest

@pytest.fixture()
def bin_dir(tmpdir: Path):
    return Path(tmpdir) / 'bin'

@pytest.fixture()
def snk_home(tmpdir: Path):
    return Path(tmpdir) / 'snk'

@pytest.fixture()
def nest(snk_home, bin_dir):
    return Nest(snk_home, bin_dir)


import gffutils

db = gffutils.create_db(
    snakemake.input[0],
    dbfn=snakemake.output.db,
    force=True,
    keep_order=True,
    merge_strategy="merge",
    sort_attribute_values=True,
    disable_infer_genes=True,
    disable_infer_transcripts=True,
)

with open(snakemake.output.bed, "w") as outfileobj:
    for tx in db.features_of_type("transcript", order_by="start"):
        bed = [s.strip() for s in db.bed12(tx).split("\t")]
        bed[3] = tx.id
        outfileobj.write("{}\n".format("\t".join(bed)))



from typing import List, Callable
from inspect import signature, Parameter
from makefun import wraps
from pathlib import Path
import typer

types = {
    'int': int,
    'integer': int,
    'str': str,
    'string': str,
    'path': Path,
    'bool': bool,
}

def create_cli_parameter(option):
    return Parameter(
        option['name'], 
        kind=Parameter.POSITIONAL_OR_KEYWORD, 
        default=typer.Option(option['default'], help=f"[CONFIG] {option['help']}"), 
        annotation=types[option['type'].lower()])

def add_dynamic_options(options: List[dict]):
    def inner(func: Callable):
        func_sig = signature(func)
        params = list(func_sig.parameters.values())
        for op in options[::-1]:
            params.insert(1, create_cli_parameter(op))
        new_sig = func_sig.replace(parameters=params)
        @wraps(func, new_sig=new_sig)
        def func_wrapper(*args, **kwargs):
            if kwargs['configfile']:
                # need to check if kwargs in options have changed
                # parse the new configfile and update the defautls
                raise NotImplementedError
            for op in options:
                kwargs['ctx'].args.extend([f"--{op['name']}", kwargs[op['name']]])
            kwargs= {k:v for k,v in kwargs.items() if k in func_sig.parameters.keys()}
            return func(*args, **kwargs)
        return func_wrapper
    return inner


import typer
from pathlib import Path
import os
from typing import Optional
from rich.pretty import pprint
from .nest import Nest
from .errors import PipelineExistsError, PipelineNotFoundError

app = typer.Typer()

SNK_HOME = None
SNK_BIN = None

@app.callback(context_settings={"help_option_names": ["-h", "--help"]})
def callback(
    home: Optional[Path] = typer.Option(
            None, 
            envvar="SNK_HOME", 
            dir_okay=True, 
            file_okay=False, 
            exists=True, 
            help="Overrides default snk location. Pipelines will be installed in $SNK_HOME/pipelines."
        ),
    bin: Optional[Path] = typer.Option(
            None, 
            envvar="SNK_BIN", 
            dir_okay=True, 
            file_okay=False, 
            exists=True, 
            help="Overrides location of pipeline installations. Pipelines are symlinked here."
        )
    ):
    """\b
        _            _             _        
       / /\         /\ \     _    /\_\      
      / /  \       /  \ \   /\_\ / / /  _   
     / / /\ \__   / /\ \ \_/ / // / /  /\_\ 
    / / /\ \___\ / / /\ \___/ // / /__/ / / 
    \ \ \ \/___// / /  \/____// /\_____/ /  
     \ \ \     / / /    / / // /\_______/   
 _    \ \ \   / / /    / / // / /\ \ \      
/_/\__/ / /  / / /    / / // / /  \ \ \     
\ \/___/ /  / / /    / / // / /    \ \ \    
 \_____\/   \/_/     \/_/ \/_/      \_\_\   
 \b
 \n
 Snakemake pipeline management system
    """
    global SNK_BIN, SNK_HOME
    SNK_BIN = bin
    SNK_HOME = home
    


@app.command()
def install(
        pipeline: str = typer.Argument(
            ..., help="URL or Github name (user/repo) of the pipeline to install."
        ),
        name: Optional[str] = typer.Option(
            None, 
            help="Rename the pipeline (this name will be used to call the CLI.)"
        ),
        version: Optional[str] = typer.Option(
            None, 
            help="Version of the pipeline to install. Can specify commit branch name, or tag. If None the latest commit will be installed."
        ),
        config: Optional[Path] = typer.Option(
            None, 
            help="Specify a non-standard config location."
        ),
    ):
    """
    Install a pipeline.
    """
    
    nest = Nest(snk_home=SNK_HOME, bin_dir=SNK_BIN)
    if not pipeline.startswith('http'):
        pipeline = f"https://github.com/{pipeline}.git"
    try:
        installl_pipeline = nest.install(repo_url=pipeline, name=name, version=version, config=config)
    except PipelineExistsError as e:
        typer.secho(e, fg='red')
        raise typer.Exit()
    except PipelineNotFoundError as e:
        typer.secho(e, fg='red')
        raise typer.Exit()
    v = installl_pipeline.version
    v = v if v else 'latest'
    typer.secho(f"Successfully installed {installl_pipeline.name} ({v})!")


@app.command()
def uninstall(
        name: str = typer.Argument(..., help="Name of the pipeline to uninstall."),
        force: Optional[bool] = typer.Option(False, '--force', '-f', help="Force uninstall without asking."),
    ):
    """
    Uninstall a pipeline.
    """
    nest = Nest(snk_home=SNK_HOME, bin_dir=SNK_BIN)
    try:
        uninstalled = nest.uninstall(name, force=force)
    except PipelineNotFoundError as e:
        typer.secho(e, fg='red')
        raise typer.Exit(1)
    if uninstalled:
        typer.secho(f"Successfully uninstalled {name}!", fg='green')
        

@app.command()
def update():
    """
    Update a pipeline.
    """
    raise NotImplementedError


@app.command()
def list():
    """
    List the installed pipelines.
    """
    nest = Nest()
    try:
        pipelines = os.listdir(nest.pipelines_dir)
    except FileNotFoundError:
        pipelines = []
    pprint(pipelines)


@app.command()
def run():
    """
    Run the pipeline in a temporary environment.
    """
    raise NotImplementedError

def annotations():
    """gen annotations defaults from config file"""

def create():
    """create a default project that can be installed with snk"""


import sys

# logging
sys.stderr = open(snakemake.log[0], "w")

import pandas as pd


def get_column(strandedness):
    if pd.isnull(strandedness) or strandedness == "none":
        return 1  # non stranded protocol
    elif strandedness == "yes":
        return 2  # 3rd column
    elif strandedness == "reverse":
        return 3  # 4th column, usually for Illumina truseq
    else:
        raise ValueError(
            (
                "'strandedness' column should be empty or have the "
                "value 'none', 'yes' or 'reverse', instead has the "
                "value {}"
            ).format(repr(strandedness))
        )


counts = [
    pd.read_table(
        f, index_col=0, usecols=[0, get_column(strandedness)], header=None, skiprows=4
    )
    for f, strandedness in zip(snakemake.input, snakemake.params.strand)
]

for t, sample in zip(counts, snakemake.params.samples):
    t.columns = [sample]

matrix = pd.concat(counts, axis=1)
matrix.index.name = "gene"
# collapse technical replicates
matrix = matrix.groupby(matrix.columns, axis=1).sum()
matrix.to_csv(snakemake.output[0], sep="\t")



from snk import Nest
from pathlib import Path
import subprocess

def test_init(bin_dir, snk_home):
    nest = Nest(snk_home, bin_dir=bin_dir)
    assert nest.pipelines_dir == Path(snk_home) / 'pipelines'
    for path in [nest.pipelines_dir, nest.bin_dir, nest.snk_home]:
        assert path.exists()

def test_download(nest: Nest):
    repo = nest.download('https://github.com/snakemake-workflows/rna-seq-star-deseq2.git', 'rna-seq-star-deseq2')
    expected_location = nest.pipelines_dir / 'rna-seq-star-deseq2'
    assert (expected_location).exists()
    assert Path(repo.git_dir).parent == expected_location

def test_create_package(nest: Nest):
    test_pipeline_path = nest.pipelines_dir / 'pipeline-name'
    test_pipeline_path.mkdir()
    path = nest.create_package(pipeline_dir=test_pipeline_path)
    assert path == test_pipeline_path / 'bin' / 'pipeline-name'

def test_link_pipeline_executable_to_bin(nest: Nest):
    pipeline_executable_path = Path('tests/data/snk/pipelines/rna-seq-star-deseq2/bin/rna-seq-star-deseq2')
    executable_path = nest.link_pipeline_executable_to_bin(pipeline_executable_path)
    assert executable_path.exists() == True and executable_path.is_symlink() == True

def test_install(nest: Nest):
    deseq2 = nest.install('https://github.com/snakemake-workflows/rna-seq-star-deseq2.git')
    expected = nest.pipelines_dir / 'rna-seq-star-deseq2'
    assert expected.exists()
    print(deseq2.executable)
    res = subprocess.run([deseq2.executable, '--help'], stdout=subprocess.PIPE , stderr=subprocess.PIPE)
    assert 'Usage: rna-seq-star-deseq2' in res.stdout.decode('utf-8')

def test_uninstall(nest:Nest):
    pass